#!/usr/bin/python3

# Copyright (C) 2018-2019, Josef Hahn
#
# This file is part of PyHtmlView.
#
# PyHtmlView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyHtmlView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyHtmlView.  If not, see <http://www.gnu.org/licenses/>.

"""
Tiny browser view without any controls, used for presenting a web application in a more native way. 
"""

import argparse
import sys


def openview(show_url=None, show_content=None, title=None, icon=None, wndclassname=None, scriptedclosing=None):
    """!
    Opens a new view with some content.
    @param show_url: The url to show content from.
    @param show_content: The static html content to show.
    @param title: The window title.
    @param icon: The window icon.
    @param wndclassname: The window class name.
    @param scriptedclosing: A javascript expression to execute for closing.
    """
    import PyQt5.QtCore as QtCore
    import PyQt5.QtGui as QtGui
    import PyQt5.QtWidgets as QtWidgets
    import PyQt5.QtWebEngineWidgets as QtWebEngineWidgets   # if changed, also change in test()!

    class WebView(QtWebEngineWidgets.QWebEngineView):
        def __init__(self, *a, **b):
            QtWebEngineWidgets.QWebEngineView.__init__(self, *a, **b)
        def contextMenuEvent(self, event):
            allowedactions = [self.page().action(QtWebEngineWidgets.QWebEnginePage.Cut),
                              self.page().action(QtWebEngineWidgets.QWebEnginePage.Copy),
                              self.page().action(QtWebEngineWidgets.QWebEnginePage.Paste),
                              self.page().action(QtWebEngineWidgets.QWebEnginePage.Undo),
                              self.page().action(QtWebEngineWidgets.QWebEnginePage.Redo),
                              self.page().action(QtWebEngineWidgets.QWebEnginePage.PasteAndMatchStyle)]
            menu = self.page().createStandardContextMenu()
            for act in menu.actions():
                if act not in allowedactions:
                    menu.removeAction(act)
            menu.exec(event.globalPos())
    class MainWindow(QtWidgets.QMainWindow):
        def __init__(self):
            QtWidgets.QMainWindow.__init__(self)
            self.forceclose = False
            self.setMinimumSize(QtCore.QSize(640, 480))
            self.resize(1200, 660)
            self.setWindowTitle(" ")
            if icon:
                qicon = QtGui.QIcon(icon)
                self.setWindowIcon(qicon)
            centralwidget = QtWidgets.QWidget(self)
            self.setCentralWidget(centralwidget)
            gridlayout = QtWidgets.QGridLayout(self)
            gridlayout.setContentsMargins(0, 0, 0, 0)
            centralwidget.setLayout(gridlayout)
            self.webview = webview = WebView(self)
            if title is None:
                webview.titleChanged.connect(lambda t: self.setWindowTitle(t))
            else:
                webview.setWindowTitle(title or " ")
            v = 3  # at early times, the 'window' object isnt fine
            def _loadstarted():
                nonlocal v
                if v > 0:
                    v -= 1
                    self.webview.page().runJavaScript("window.pyhtmlview=true")
            webview.loadProgress.connect(_loadstarted)
            gridlayout.addWidget(webview, 0, 0)
            if show_url:
                webview.load(QtCore.QUrl(show_url))
            elif show_content:
                webview.setHtml(show_content)
        def closeEvent(self, event):
            if scriptedclosing and not self.forceclose:
                event.ignore()
                def oncloseresult(res):
                    if res:
                        self.forceclose = True
                        self.close()
                self.webview.page().runJavaScript(scriptedclosing, oncloseresult)
            else:
                event.accept()

    app = QtWidgets.QApplication(sys.argv)
    if wndclassname:
        app.setApplicationName(wndclassname)
    mainwnd = MainWindow()
    mainwnd.show()
    app.exec_()
    return 0


def test():
    """!
    Tests if your system is compatible with PyHtmlView.
    
    If not, it throws an exception.
    """
    import PyQt5.QtWebEngineWidgets
    return 0


def _execute(args):
    fct = globals().get(str(args.command), None)
    if fct:
        return fct(**{k: v for k, v in args.__dict__.items() if k != "command"})
    else:
        raise Exception("Please specify a valid command. Use '--help' for more.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="PyHtmlView tiny browser view.")
    cmds = parser.add_subparsers(help='The command to execute.', metavar="command", dest="command")
    c_openview = cmds.add_parser('openview', help='Opens a new view.')
    c_openview.add_argument('--show-url', help='The url to show content from.')
    c_openview.add_argument('--show-content', help='The static html content to show.')
    c_openview.add_argument('--title', help='Sets the window title.')
    c_openview.add_argument('--icon', help='Sets the window icon (path to a png file).')
    c_openview.add_argument('--wndclassname', help='Sets the window class name (used by the window manager).')
    c_openview.add_argument('--scriptedclosing', help='Sets a javascript expression to execute for closing.')
    c_test = cmds.add_parser('test', help='Tests if your system is compatible with PyHtmlView (return code 0: yes).')
    cmdargs = parser.parse_args()
    try:
        sys.exit(_execute(cmdargs))
    except Exception as e:
        print("Error:", str(e))
        sys.exit(1)
