# Copyright (C) 2014-2019, Josef Hahn
#
# This file is part of PyWebApp.
#
# PyWebApp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyWebApp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyWebApp.  If not, see <http://www.gnu.org/licenses/>.

"""
Web-based user interfaces running in a local browser.
"""

import datetime
import errno
import http.server
import json
import os
import re
import shutil
import socketserver
import subprocess
import sys
import threading
import time
import traceback
import urllib.parse
import urllib.request
import weakref


basedir = os.path.dirname(os.path.abspath(__file__))


def do_async(fct):
    def sfct(self, request, **a):
        def afct(_request, **_a):
            if fct(self, _request, **_a):
                raise Exception("async handlers must not return stuff")
        threading.Thread(target=request._requesthandler._callhandler,
                         args=(afct, request._params, request._headers)).start()
    return sfct


def do_with_param_type(paramname, paramtype):
    def _fct(fct):
        def _f_fct(self, request, **a):
            if paramname in a:
                try:
                    a[paramname] = paramtype(a[paramname])
                except:
                    pass  # then we leave it as it is; maybe it works or not ;)
            return fct(self, request, **a)
        return _f_fct
    return _fct


# noinspection PyUnusedLocal
class Application:
    """!
    Base class for a PyWebApp web application.
    """

    _idcnt = 0
    
    def __init__(self, parentid=None, returntoparent=True):
        self.parentid = parentid
        self.returntoparent = returntoparent
        self._nextbrowseropenusetakeover = bool(parentid)
        self.id = "{}n{}".format(os.getpid(), Application._idcnt)
        Application._idcnt += 1
        self.url = None
        self.takenover = False
        self.wetookoverfromurl = None
        self.wetookoverfromurl_last = None
        self.httpd = None
        self._stopping = False
        self.browserhook = None
        self.lastopenedbrowserpid = None
        self.appconfig = Application.AppConfig(self)
        self.appconfig.setconfig("showBrowserClosedNotification", False)
        self.appconfig.setconfig("stopImplicitlyWhenBrowserClosed", False)
        self.appconfig.setconfig("AbstractFilesystemDialog_showChangeDriveButton", sys.platform == "win32")
        self.userfeedback = Application.UserFeedbackController(self)
        self._events = []
        self._event_lastid = -1
        self._eventcondition = threading.Condition()
        self._datastores = {}
        self._wasreopened = False
        self._clientscript = b""
        self._clientstyle = b""
        self._pyhtmlview_additionalargs = []
        self._pyhtmlview_enabled = None

    def create_datastore(self, name):
        """!
        Creates an Application.DataStore with a name and returns it.
        """
        self._datastores[name] = result = Application.DataStore(name, self)
        return result

    def start(self):
        """!
        Starts the application.
        """
        if self.httpd:
            return
        self.oninitialize()
        port = 7309
        connected = False
        while not connected:
            try:
                self.httpd = socketserver.ThreadingTCPServer(("", port), Application.PyWebAppHTTPRequestHandler)
                self.httpd.webmodule = self
                connected = True
            except OSError as e:
                if e.errno == errno.EADDRINUSE:
                    port += 1
                else:
                    raise
        self.url = "http://localhost:{}/".format(port)
        os.environ["PYWEBAPP_" + self.id + "_URL"] = self.url
        self.browserhook = Application.BrowserHook(self.url, self)
        self.browserhook.start()
        threading.Thread(target=self.httpd.serve_forever, daemon=False).start()

    def stop(self):
        """!
        Stops the application.
        """
        if not self.httpd:
            return

        def _stop():
            self.browserhook.stop()
            self._stopping = True
            for i in range(10):
                if not self.takenover:
                    break
                time.sleep(1)
            if self.wetookoverfromurl:
                try:
                    turl = self.wetookoverfromurl + "returntakeover"
                    with urllib.request.urlopen(turl) as f:
                        f.read()
                except OSError:
                    pass
            self.triggerevent("stopapplication", targeturl=self.wetookoverfromurl if self.returntoparent else None)
            self.wetookoverfromurl = None
            norequestcnt = 0
            for i in range(120):
                time.sleep(0.25)
                with Application.PyWebAppHTTPRequestHandler.openrequestcnt_lock:
                    reqs = Application.PyWebAppHTTPRequestHandler.openrequestcnt
                if reqs == 0:
                    if norequestcnt == 12:
                        break
                    norequestcnt += 1
                else:
                    norequestcnt = 0
            self.httpd.shutdown()
            self.httpd.socket.close()
            self._stopping = False
        threading.Thread(target=_stop).start()
        
    def waituntilstopped(self):
        """!
        Blocks execution while the application is running.
        """
        while self.isrunning():
            time.sleep(1)

    def isrunning(self):
        """!
        Checks if the application is currently running (started and not in shutdown process).
        """
        return (self.httpd is not None) and (not self._stopping)

    def staticcontent(self):
        """!
        Returns the main page content.
        @a_infrastructure
        """
        with open(basedir + "/pywebapp.html", "rb") as f:
            cnt = f.read()
        return cnt.replace(b"{{ additionalstyle }}", self._clientstyle)

    def oninitialize(self):
        """!
        Initializes the application.
        @a_dooverride_optional
        """
        pass

    def storeasstaticapplication(self, targetdir):
        """!
        Stores the application as static files.
        """
        self.oninitialize()
        self.httpd = object()  # otherwise it is not considered as 'running' internally
        self.appconfig.setconfig("isStaticApplication", True)
        os.makedirs(targetdir+"/static")
        with open(targetdir + "/index.html", "wb") as f:
            f.write(self.staticcontent())
        with open(targetdir + "/pywebapp_init", "wb") as f:
            f.write(self.do_pywebapp_init(None))
        for sf in os.listdir(basedir):
            fsf = basedir + "/" + sf
            if os.path.isfile(fsf):
                shutil.copyfile(fsf, targetdir+"/static/"+sf)
    
    def onprocessrequesterror(self, request, error):
        """!
        Reacts on process request errors.
        @a_dooverride_optional
        """
        print("Program error in processing request '{}': {}".format(request, traceback.format_exc()))

    def onopenbrowsererror(self):
        """!
        Reacts on errors while opening the browser.
        @a_dooverride_optional
        """
        print("It was impossible to open a browser for showing the application.\nPlease open a browser manually and go "
              "to the application address (see above).")

    def onunhandledclienterror(self, error, src, line):
        """!
        Reacts on unhandled client errors.
        @a_dooverride_optional
        """
        errortext = str(error)
        src = src or ""
        if line:
            src += " :" + str(line)
        if src:
            errortext += "\n(" + src + ")"
        self.triggerevent("showunhandlederror", errortext=errortext)
        return False

    def onopenbrowserinformationoutput(self, kind, message):
        """!
        Reacts on information output while opening the browser.
        @a_dooverride_optional
        """
        print(message)

    def onbrowserreopened(self):
        """!
        Reacts on the fact that the browser was opened again.
        @a_dooverride_optional
        """
        pass

    def tryclosebrowser(self, withchecks=True):
        """!
        Tries to close the browser window.
        
        Note: This will not work in all situations, e.g. typically not if the app is opened in the system browser.
        Your application should handle that situation gracefully (e.g. by the default app shutdown notification).
        """
        if withchecks:
            if not (self.appconfig.getconfig("stopImplicitlyWhenBrowserClosed") or
                    self.appconfig.getconfig("showBrowserClosedNotification")):
                return
        success = False
        if self._pyhtmlview_enabled and self.lastopenedbrowserpid:
            try:
                os.kill(self.lastopenedbrowserpid, 15)
            except OSError:
                pass
            success = True
        elif self.wetookoverfromurl_last:
            try:
                with urllib.request.urlopen(self.wetookoverfromurl_last + "application_tryclosebrowser") as f:
                    tosucc = f.read()
            except Exception:
                tosucc = ""
            return tosucc == b"true"
        return success

    def openbrowser(self, url):
        """!
        Opens a browser view.
        @a_infrastructure
        """
        self._pyhtmlview_enabled = False
        self.lastopenedbrowserpid = 0
        self.onopenbrowserinformationoutput(0, "Your application runs at url {}.".format(url))
        if self._nextbrowseropenusetakeover:
            self._nextbrowseropenusetakeover = False
            try:
                self.wetookoverfromurl = self.wetookoverfromurl_last = os.environ["PYWEBAPP_" + self.parentid + "_URL"]
                turl = self.wetookoverfromurl + "takeover?url=" + urllib.parse.quote_plus(self.url)
                with urllib.request.urlopen(turl) as f:
                    f.read()
                return
            except OSError:
                pass
        cmdline = os.getenv("PYWEBAPP_CALLCMD")
        if not cmdline:
            # we use PyHtmlView if available and usable
            pyhtmlviewpath = os.path.abspath(basedir + "/../PyHtmlView/pyhtmlview.py")
            if os.path.isfile(pyhtmlviewpath):
                ret = subprocess.call([pyhtmlviewpath, "test"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                if ret == 0:
                    cmdline = [pyhtmlviewpath, "openview", "--show-url", "%u",
                               "--wndclassname", type(self).__name__,
                               "--scriptedclosing", "pywebapp.pyhtmlviewsupport._onwindowclose()"
                               ] + self._pyhtmlview_additionalargs
                    self._pyhtmlview_enabled = True
        if not cmdline:
            if sys.platform == "win32":
                # noinspection PyUnresolvedReferences
                threading.Thread(target=lambda: os.startfile(url)).start()
            else:
                try:
                    subprocess.check_output(["xdg-open --version"], shell=True)
                    cmdline = "nohup xdg-open %u"
                except subprocess.CalledProcessError:
                    self.onopenbrowserinformationoutput(
                        1, "Please start a browser and navigate to the application.".format(url))
        if cmdline:
            def _open():
                nonlocal cmdline
                cmdline_shell = False
                if isinstance(cmdline, str):
                    cmdline = [cmdline]
                    cmdline_shell = True
                cmdline = [c.replace("%u", url) for c in cmdline]
                with subprocess.Popen(cmdline, shell=cmdline_shell, stdout=subprocess.DEVNULL,
                                      stderr=subprocess.DEVNULL) as bp:
                    self.lastopenedbrowserpid = bp.pid
                    bp.wait()
            threading.Thread(target=_open, daemon=True).start()

    @staticmethod
    def _get_content(content, frompath):
        result = b""
        if isinstance(content, str):
            result += content.encode() + b"\n"
        elif isinstance(content, bytes):
            result += content + b"\n"
        elif content is not None:
            result += str(content).encode() + b"\n"
        if frompath is not None:
            with open(frompath, "rb") as f:
                result += f.read() + b"\n"
        return result

    def addclientstyle(self, content=None, *, frompath=None):
        """!
        Adds additional css application styles.

        Call this from your constructor.
        """
        self._clientstyle += self._get_content(content, frompath)

    def addclientscript(self, content=None, *, frompath=None):
        """!
        Adds application client javascript.

        Call this from your constructor.
        """
        self._clientscript += self._get_content(content, frompath)

    def clientstaticfilelocations(self):
        """!
        Returns additional locations for static files.
        @a_dooverride_optional
        """
        return []

    def triggerevent(self, name, **eventdata):
        """!
        Triggers an event on client side.
        """
        with self._eventcondition:
            eventid = self._event_lastid + 1
            self._event_lastid = eventid
            eventdata.update({'name': name})
            self._events.append((eventid, datetime.datetime.now(), eventdata))
            
    def do_pywebapp_init(self, request):
        with self._eventcondition:    
            appconfigs = self.appconfig.getconfigs(markinitialized=True)
            wasreopened = self._wasreopened
            self._wasreopened = False
            return self._clientscript + \
                "\npywebapp.initialize({running}, {wasreopened}, {eventlastid}, {appconfigs});".format(
                    running=json.dumps(self.isrunning()), eventlastid=json.dumps(self._event_lastid),
                    appconfigs=json.dumps(appconfigs, default=Application.json_make_serializable),
                    wasreopened=json.dumps(wasreopened)).encode()

    @do_with_param_type("lastid", int)
    def do_pullevent(self, request, lastid):
        self.browserhook.heartbeat()
        begintime = datetime.datetime.now()
        while (datetime.datetime.now() - begintime).total_seconds() < 3:
            result = []
            for eid, etime, eobject in self._events:
                if eid > lastid:
                    result.append({'id': eid, 'data': eobject})
            if len(result) > 0:
                return result
            with self._eventcondition:
                self._eventcondition.wait(1)
        return {}

    @do_with_param_type("id", int)
    def do_datastore_pull(self, request, name, id=None):
        datastore = self._datastores[name]
        if id is not None:
            cell = Application.DataStore.getcellbyid(id)
            if cell is None:
                return None
            else:
                ptr = datastore.cell2ptr(cell)
        else:
            ptr = None
            cell = datastore.rootcell
        rows = []
        for irow in range(datastore.rowcount(ptr)):
            cols = []
            for icol in range(datastore.columncount(ptr)):
                icell = datastore.ptr2cell(datastore.valuepointer(irow, icol, ptr))
                cols.append((icell.ajaxid, icell.value))
            rows.append(cols)
        return {"value": cell.value, "childs": rows, "id": cell.ajaxid}

    def do_datastore_push(self, request):
        pass  # not yet supported

    def do_application_tryclosebrowser(self, request, checks="1"):
        return self.tryclosebrowser(checks)

    def do_application_shutdown(self, request):
        self.stop()

    def do_unhandled_client_error(self, request, error="", src="", line=""):
        shutdown = self.onunhandledclienterror(error, src, line)
        return {'shutdown': bool(shutdown)}

    def do_answer_userfeedback(self, request, **a):
        # noinspection PyProtectedMember
        self.userfeedback._interpret_response(request)

    def do_listfs(self, request, path, onlyDirs, showHidden):
        dirs = []
        if sys.platform == "win32" and path == "":
            import string
            from ctypes import windll
            drivesmask = windll.kernel32.GetLogicalDrives()
            for letter in string.ascii_uppercase:
                if drivesmask & 1:
                    dirs.append((False, letter + ":"))
                drivesmask >>= 1
        else:
            if sys.platform == "win32" and re.match("/([A-Z]):", path):
                path = path[1:]
            path = os.path.abspath(path)
            try:
                for x in os.listdir(path):
                    isfile = os.path.isfile(path+"/"+x)
                    if not (onlyDirs and isfile):
                        if showHidden or (not x.startswith(".")):
                            dirs.append((isfile, x))
            except IOError:
                pass
            dirs = sorted(dirs)
            if path != os.path.dirname(path):
                dirs.insert(0, (False, ".."))
        return {'path': path, 'dirs': dirs}

    def do_setappconfigvalue(self, request, key, value):
        self.appconfig.setconfig(key, json.loads(value))

    def do_takeover(self, request, url):
        self.takenover = True
        self.appconfig.setconfig("takeoverurl", url)

    def do_returntakeover(self, request):
        self.takenover = False
        self.appconfig.setconfig("takeoverurl", None)

    def sethead1(self, s):
        """!
        Sets the first head string.
        """
        self.appconfig.setconfig("head1", s)
    
    def sethead2(self, s):
        """!
        Sets the second head string.
        """
        self.appconfig.setconfig("head2", s)

    def seticon(self, src):
        """!
        Sets the icon source.
        """
        self.appconfig.setconfig("icon", src)

    class SetStopImplicitlyWhenBrowserClosed:
        """!
        @a_infrastructure
        See stop_implicitly_when_browser_closed() and dont_stop_implicitly_when_browser_closed().
        """

        def __init__(self, app, val):
            self.app = app
            self.orig = False
            self.val = val

        def __enter__(self):
            self.orig = self.app.appconfig.getconfig("stopImplicitlyWhenBrowserClosed")
            self.app.appconfig.setconfig("stopImplicitlyWhenBrowserClosed", self.val)
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.app.appconfig.setconfig("stopImplicitlyWhenBrowserClosed", self.orig)

    def stop_implicitly_when_browser_closed(self):
        """!
        Use such instances via `with` for temporarily enabling implicit application stop when browser closed.
        """
        return Application.SetStopImplicitlyWhenBrowserClosed(self, True)

    def dont_stop_implicitly_when_browser_closed(self):
        """!
        Use such instances via `with` for temporarily disabling implicit application stop when browser closed.
        """
        return Application.SetStopImplicitlyWhenBrowserClosed(self, False)

    class UserFeedbackController:
        """!
        A controller for user feedback operations.

        Use methods inside it for communicating with the user during some operations.
        Such an object is automatically available during execution, there is no need to create new ones.
        """

        def __init__(self, app):
            self.nextreqid = 0
            self._responses = {}
            self.app = app
            self._responseslock = threading.Lock()
            self._responsescond = threading.Condition(self._responseslock)
            self.nix = object()

        def make_raw_request(self, kind, **params):
            reqid = self.nextreqid
            self.nextreqid += 1
            r = {'reqid': reqid, 'kind': kind}
            r.update(params)
            self.app.appconfig.setconfig("UserFeedback_currentfeedback", r)
            self._responses[reqid] = self.nix
            with self._responseslock:
                while True:
                    answer = self._responses[reqid]
                    if answer != self.nix:
                        break
                    self._responsescond.wait()
                self._responses.pop(reqid)
            self.app.appconfig.setconfig("UserFeedback_currentfeedback", None)
            return answer

        def _interpret_response(self, request):
            reqid = int(request.gparam("reqid"))
            with self._responseslock:
                if self._responses[reqid] == self.nix:
                    self._responses[reqid] = json.loads(request.gparam("answer"))[0]
                    self._responsescond.notify_all()

        def messagedialog(self, message, buttons=None):
            if buttons is None:
                buttons = [Application.ClientString("OK")]
            return int(self.make_raw_request("message", message=message, buttons=buttons))

        def inputdialog(self, question, defaultanswer=""):
            return self.make_raw_request("input", question=question, defaultanswer=defaultanswer)

        def multilineinputdialog(self, question, defaultanswer=""):
            return self.make_raw_request("multilineinput", question=question, defaultanswer=defaultanswer)

        def filesystemdialog(self, fstype="file", question="", startpath="/"):
            return self.make_raw_request("filesystem", fstype=fstype, question=question, startpath=startpath)

    class AppConfig:
        """!
        Application configuration storage.

        This can store arbitrary configuration items, which then become available on client side as well.
        """

        def __init__(self, app):
            self.cfg = {}
            self.app = app
            self._appinitialized = False
            self.lock = threading.Lock()

        def markinitialized(self):
            self._appinitialized = True

        def getconfigs(self, markinitialized=False):
            with self.lock:
                if markinitialized:
                    self.markinitialized()
                return dict(self.cfg)

        def getconfig(self, key):
            with self.lock:
                return self.cfg[key]

        def setconfig(self, key, value):
            with self.lock:
                self.cfg[key] = value
                if self._appinitialized:
                    self.app.triggerevent("AppConfig_setconfig", key=key, value=value)

    class DataStore:
        """!
        A clove-like datastore which then becomes available on client side as well.
        """

        _lock = threading.Lock()
        _nextid = 0
        _celldict = {}

        @staticmethod
        def getcellbyid(ajaxid):
            with Application.DataStore._lock:
                if ajaxid not in Application.DataStore._celldict:
                    return None
                x = Application.DataStore._celldict[ajaxid]()
                if x is None:
                    Application.DataStore._celldict.pop(ajaxid)
                return x

        class ValuePointer:

            def __init__(self, irow, icol, cell):
                self.irow = irow
                self.icol = icol
                self.cell = cell

        class Cell:

            def __init__(self, datastore):
                with Application.DataStore._lock:
                    self.ajaxid = Application.DataStore._nextid
                    Application.DataStore._nextid += 1
                    Application.DataStore._celldict[self.ajaxid] = weakref.ref(self)
                self.datastore = datastore
                self.value = None
                self.irow = 0
                self.icol = 0
                self.rowcount = 0
                self.colcount = 0
                self.childs = []
                self.parent = None

        def _triggerdsevent(self, cell):
            self.app.triggerevent("DataStore_update", datasource=self.name, id=self._cellids(cell))

        def __init__(self, name, app):
            self.name = name
            self.app = app
            self.rootcell = Application.DataStore.Cell(self)

        def rowcount(self, ptr=None):
            return self.ptr2cell(ptr).rowcount

        def columncount(self, ptr=None):
            return self.ptr2cell(ptr).colcount

        def insertrow(self, i, ptr=None, values=None):
            if values is None:
                values = []
            pcell = self.ptr2cell(ptr)
            pcell.childs.insert(i, [])
            for j in range(i+1, pcell.rowcount):
                for jc in pcell.childs[j]:
                    jc.irow = j
            pcell.rowcount += 1
            for jv in range(len(values)):
                if self.columncount(ptr) <= jv:
                    self.insertcolumn(jv, ptr)
                self.setvalue(self.valuepointer(i, jv, ptr), values[jv])
            self._triggerdsevent(pcell)

        def insertcolumn(self, i, ptr=None, values=None):
            if values is None:
                values = []
            pcell = self.ptr2cell(ptr)
            for jr in pcell.childs:
                if i < len(jr):
                    jr.insert(i, None)
                    for j in range(i+1, len(jr)):
                        jr[j].icol = j
            pcell.colcount += 1
            for jv in range(len(values)):
                if self.rowcount(ptr) <= jv:
                    self.insertrow(jv, ptr)
                self.setvalue(self.valuepointer(jv, i, ptr), values[jv])
            self._triggerdsevent(pcell)

        def removerow(self, i, ptr=None):
            pcell = self.ptr2cell(ptr)
            pcell.childs.pop(i)
            for j in range(i, pcell.rowcount):
                for jc in pcell.childs[j]:
                    jc.irow = j
            pcell.rowcount -= 1
            self._triggerdsevent(pcell)

        def removecolumn(self, i, ptr=None):
            pcell = self.ptr2cell(ptr)
            for jr in pcell.childs:
                if i < len(jr):
                    jr.pop(i)
                    for j in range(i, len(jr)):
                        jr[j].icol = j
            pcell.colcount -= 1
            self._triggerdsevent(pcell)

        def getvalue(self, ptr=None):
            return self.ptr2cell(ptr).value

        def _cellids(self, cell):
            result = []
            while cell != self.rootcell:
                result.append(cell.ajaxid)
                cell = cell.parent
            return result

        def setvalue(self, ptr=None, value=None):
            cell = self.ptr2cell(ptr)
            cell.value = value
            self._triggerdsevent(cell)

        def parent(self, ptr=None):
            pcell = self.ptr2cell(ptr)
            return Application.DataStore.ValuePointer(pcell.irow, pcell.icol, pcell)

        def ptr2cell(self, ptr):
            return ptr.cell if ptr else self.rootcell

        def cell2ptr(self, cell):
            result = None
            cpath = []
            ccell = cell
            while ccell != self.rootcell:
                cpath.append(ccell)
                ccell = ccell.parent
            for dcell in cpath:
                result = self.valuepointer(dcell.irow, dcell.icol, result)
            return result

        def valuepointer(self, irow, icol, parent):
            pcell = self.ptr2cell(parent)
            if irow >= pcell.rowcount or icol >= pcell.colcount:
                raise Exception("bad pointer")
            row = pcell.childs[irow]
            while len(row) <= icol:
                row.append(None)
            if row[icol] is None:
                row[icol] = Application.DataStore.Cell(self)
            cell = row[icol]
            cell.parent = pcell
            cell.irow = irow
            cell.icol = icol
            return Application.DataStore.ValuePointer(irow, icol, cell)

    class BrowserHook:
        """!
        Helper for opening browsers.
        @a_infrastructure
        """

        def __init__(self, url, app):
            self.url = url
            self.lastheartbeat = None
            self.app = app
            self._stop = False
            self._triedforheartbeatx = 0
            self._heartbeatx = None

        def heartbeat(self):
            self.lastheartbeat = datetime.datetime.now()

        def start(self):
            def _reopenerthread():
                self._stop = False
                while not self._stop:
                    if self.app.takenover:
                        while self.app.takenover:
                            time.sleep(1)
                        time.sleep(30)
                    # noinspection PyTypeChecker
                    if (self.lastheartbeat is None) \
                            or (datetime.datetime.now()-self.lastheartbeat).total_seconds() > 30:
                        if self._heartbeatx != self.lastheartbeat:
                            self._heartbeatx = self.lastheartbeat
                            self._triedforheartbeatx = 1
                        elif self._triedforheartbeatx > 1:
                            self.app.onopenbrowsererror()
                            break
                        else:
                            self._triedforheartbeatx += 1
                        if self.lastheartbeat:
                            if self.app.appconfig.getconfig("stopImplicitlyWhenBrowserClosed"):
                                self.app.stop()
                                break
                            self.app._wasreopened = True
                            self.app.onbrowserreopened()
                        self.app.openbrowser(self.url)
                        time.sleep(30)
                    time.sleep(1)
            threading.Thread(target=_reopenerthread, daemon=True).start()

        def stop(self):
            self._stop = True

    class ClientRequest:
        """!
        A request from the client.
        """

        def __init__(self, handler, params, headers, requesthandler):
            self._handler = handler
            self._params = params
            self._fparams = {x[0]: x[1][0] for x in params.items()}
            self._headers = headers
            self._requesthandler = requesthandler

        def execute(self):
            return self._handler(self, **self._fparams)

        def gparam(self, name, defaultval=None):
            return self._fparams.get(name, defaultval)

    class ClientString:
        """!
        A i18n-aware string for displaying on client side. When returned in a json structure, the client side
        transparently gets a translated version.
        """

        def __init__(self, stringname, **args):
            self.stringname = stringname
            self.args = args

    @staticmethod
    def json_make_serializable(o):
        """!
        Makes some special objects json serializable.
        @a_infrastructure
        """
        if isinstance(o, Application.ClientString):
            return {'_pywebapp_clientstring': True, 'stringname': o.stringname, 'args': o.args}
        else:
            raise TypeError()

    class PyWebAppHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
        """!
        The http request handler.
        @a_infrastructure
        """

        openrequestcnt_lock = threading.Lock()
        openrequestcnt = 0

        def log_message(self, sformat, *args):
            pass

        def do_HEAD(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

        # noinspection PyPep8Naming
        def do_POST(self):
            return self.do_GET()

        def _callhandler(self, handler, params, headers):
            try:
                return Application.ClientRequest(handler, params, headers, self).execute(), None, None
            except Exception as e:
                handled = self.server.webmodule.onprocessrequesterror(self.path, e)
                if not handled:
                    self.server.webmodule.triggerevent("showunhandlederror", errortext=traceback.format_exc())
                return None, e, traceback.format_exc()

        def do_GET(self):
            with Application.PyWebAppHTTPRequestHandler.openrequestcnt_lock:
                Application.PyWebAppHTTPRequestHandler.openrequestcnt += 1
            try:
                pourl = urllib.parse.urlparse(self.path)
                params = dict(urllib.parse.parse_qs(pourl.query, keep_blank_values=True))
                if 'Content-Length' in self.headers:
                    data_string = self.rfile.read(int(self.headers['Content-Length']))
                    params.update(urllib.parse.parse_qs(data_string.decode(), keep_blank_values=True))
                # generate answer
                headers = {}
                errortxt = None
                if self.path == "/":
                    headers["Content-type"] = "text/html;charset=utf-8"
                    body = self.server.webmodule.staticcontent()
                elif self.path.startswith("/static/"):
                    for bdir in ([basedir]+self.server.webmodule.clientstaticfilelocations()):
                        bfile = bdir + self.path[7:]
                        if os.path.isfile(bfile):
                            with open(bfile, "rb") as f:
                                cnt = f.read()
                            body = cnt
                            break
                else:
                    headers["Content-type"] = "text/javascript;charset=utf-8"
                    handler = getattr(self.server.webmodule, "do_"+pourl.path[1:], None)
                    if handler:
                        result, _exception, errortxt = self._callhandler(handler, params, headers)
                    else:
                        result = None
                    body = result if isinstance(result, bytes) \
                        else json.dumps(result, default=Application.json_make_serializable).encode()
                # send answer
                if errortxt:
                    self.send_error(500, explain=errortxt)
                else:
                    self.send_response(200)
                    for headerkey, headervalue in headers.items():
                        self.send_header(headerkey, headervalue)
                    self.end_headers()
                    try:
                        self.wfile.write(body)
                    except BrokenPipeError:
                        pass
            finally:
                with Application.PyWebAppHTTPRequestHandler.openrequestcnt_lock:
                    Application.PyWebAppHTTPRequestHandler.openrequestcnt -= 1


if __name__ == "__main__":
    
    class SampleApplication(Application):
        
        def __init__(self):
            Application.__init__(self)
            self.clockdata = self.create_datastore("clockdata")
            self.sethead1("Clock")
            self.seticon("⏲")

            def _updateclock():
                while True:
                    self.clockdata.setvalue(value=datetime.datetime.now().strftime("%H:%M:%S"))
                    time.sleep(0.5)
            threading.Thread(target=_updateclock, daemon=True).start()
            self.addclientscript(self.clientscript())

        def oninitialize(self):
            self.appconfig.setconfig("stopImplicitlyWhenBrowserClosed", True)

        def clientscript(self):
            return """
                pywebapp.ready(function(){
                    var clockdata = pywebapp.getDatasource("clockdata");
                    pywebapp.mainview.setBodyLeft({rows:[{view: 'Spacer', verticalStretchAffinity: 1}, 
                                                         {label: clove.databind({datasource: clockdata}), 
                                                          style: "font-size:5em;letter-spacing:3pt;text-align:center;", 
                                                          horizontalStretchAffinity: 1}, 
                                                         {view: 'Spacer', verticalStretchAffinity: 1}], 
                                                        horizontalStretchAffinity: 1});
                    pywebapp.mainview.setActions([{name: 'close'}, {name: 'info'}]);
                    pywebapp.mainview.OnActionTriggered.addHandler(function(e){
                        if (e.action == "close")
                            pywebapp.shutdown();
                        else if (e.action == "info")
                            clove.messageDialog({message: "This is the PyWebApp sample application."});
                    });
                });
                """
    
    SampleApplication().start()
