# Copyright (C) 2014-2019, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for specifying, generating and processing all kinds of project documentation.
"""

import colorsys
import math
import os
import re
import shlex
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.documentation.exports = internals.Exports()
        universe.documentation.pool = internals.Pool()
        universe.documentation.snippets = internals.ShortSnippets()
        universe.documentation.readme_include_developer_doc = False
        universe.documentation.generate_readme = True

        def _writedocumentations(destdir):
            with universe.executionscope.newsubscope("Adding documentation"):
                internals.writedocumentations(destdir=destdir, throw_on_existing=True, fortype=PackageExport)
        universe.addhook(anise.framework.features.loadfeature("packages").featuremodule.HOOK_ENRICHRAWPACKAGE,
                         _writedocumentations)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION,
                                          provides=["_compose_readmes"], requires=["base"])
    def _compose_readmes():

        if not universe.documentation.generate_readme:
            return

        def token_in_map(t, m):
            for mk, ms in m:
                if mk == t:
                    return ms
            return None

        maturity = universe.base.getprojectstatusdescription()
        dincludedeveloperdoc = universe.documentation.readme_include_developer_doc
        dkwargs = {}
        if isinstance(dincludedeveloperdoc, dict):
            dkwargs = dincludedeveloperdoc
            dincludedeveloperdoc = True
        readmedefaultloc = universe.specialpaths.projectdir + "/_meta/readme.src"
        if os.path.isfile(readmedefaultloc):
            readmetxt = anise.utils.basic.readfromfile(readmedefaultloc, onestring=True)
            universe.documentation.exports.add(HomepageExport(key="FULLREADME", heading="Manual",
                                                              includedeveloperdoc=dincludedeveloperdoc, **dkwargs))
        else:
            readmetxt = ""
        if not token_in_map("FULLREADME", universe.documentation.pool.list):
            _lauthors = getattr(universe.base, "authors", [])
            _authors = " and ".join(([", ".join(_lauthors[0:-1])] if len(_lauthors) > 1 else []) + _lauthors[-1:])
            if getattr(universe.base, "longdescription", ""):
                universe.documentation.pool.add("_LONGDESCRIPTION", universe.base.longdescription)
                _slongdesc = "# About {#h_about}\n\\copydoc _LONGDESCRIPTION"
            else:
                _slongdesc = ""

            def fulltxt():
                _sversion = getattr(universe.packages, "getversion", lambda: None)()
                if _sversion:
                    _sversion = "You are currently reading the manual for version {}.".format(_sversion)
                else:
                    _sversion = ""
                if universe.base.license:
                    _slicense = " under the terms of the {}".format(universe.base.license.name)
                else:
                    _slicense = ""
                return """
\\tableofcontents

# License {lb}#h_license{rb}
{projectname} is written by {authors}{license}.

Please read the `LICENSE` file from the package and the \\ref h_dependencies "Dependencies" section for included
third-party stuff.

{about}

# Up-to-date?
Are you currently reading from another source than the homepage? Are you in doubt if that place
is up-to-date? If yes, you should visit {thehomepage} and check that. {sversion}

# Maturity {lb}#h_maturity{rb}
{maturity}

\\anchor _dependencies
# Dependencies {lb}#h_dependencies{rb}
<div class='deps'>
{dependencies}
</div>

{readmetxt}
                    """.format(lb="{", rb="}", projectname=universe.name, sversion=_sversion,
                               dependencies=universe.dependencies.getdependencydocument(), about=_slongdesc,
                               thehomepage=universe.base.websiteurl or "the homepage",
                               authors=_authors, license=_slicense, maturity=maturity, readmetxt=readmetxt)

            universe.documentation.pool.add("FULLREADME", fulltxt)
        heading = universe.name[0].upper() + universe.name[1:] + " Manual"
        if not os.path.exists(universe.specialpaths.projectdir+"/README"):
            universe.documentation.exports.add(PackageExport(key="FULLREADME", outputfile="/README",
                                                             format=ExportFormat.Plaintext, heading=heading,
                                                             includedeveloperdoc=dincludedeveloperdoc, **dkwargs))
        if not os.path.exists(universe.specialpaths.projectdir+"/README.pdf"):
            universe.documentation.exports.add(PackageExport(key="FULLREADME", outputfile="/README.pdf",
                                                             format=ExportFormat.PDF, heading=heading,
                                                             includedeveloperdoc=dincludedeveloperdoc, **dkwargs))

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION, requires=["_compose_readmes"])
    def _add_homepage_section():

        def my_homepage_section(homepage, sourcedir, version, head4, bannerimage, *kwargs):
            documentation = "The following documentation is available:\n\n"
            hpexports = [x for x in universe.documentation.exports.list if type(x) is HomepageExport]
            hpexports = sorted(hpexports, key=lambda x: x.index)
            for echan in hpexports:
                os.mkdir(homepage.path() + "/" + echan.key)
                elinktarget = echan.linktarget if echan.linktarget else (echan.key + ".html")
                echan.source = anise.framework.files.Filestructure(sourcedir)
                echan.generate(homepage.path() + "/" + echan.key)
                documentation += "[{title}]({ekey}/{elinktarget})\n\n".format(title=echan.heading, ekey=echan.key,
                                                                              elinktarget=elinktarget)
            if len(hpexports) > 0:
                return documentation

        universe.homepage.sections.add(3000, "documentation", "Documentation", my_homepage_section)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_release_task_for_export_to_source():
        if len([x for x in universe.documentation.exports.list if isinstance(x, SourceExport)]) > 0:
            def create_documentations_in_source():
                internals.writedocumentations(universe.specialpaths.projectdir, throw_on_existing=False,
                                              fortype=SourceExport)
            universe.releasing.tasks.add(create_documentations_in_source)

    class Pool:
        """!
        Storage of all documentation text generators.
        """

        def __init__(self):
            self.list = []

        def add(self, key, src):
            """!
            Adds a documentation.
            @param key: The documentation key for storage (later used for referencing this text).
            @param src: The documentation source text (or a function which returns one).
            """
            self.list.append([key, src])

        def gettexts(self):
            """!
            Returns all texts.
            @return: A list of key/value pairs. Both are always strings.
            """
            return [[x[0], (x[1]() if callable(x[1]) else x[1])] for x in self.list]

        def load_anisedoc_source(self, src):
            """!
            Loads the content of anisedoc sources (which contain ANISEDOC tags) into the documentation store.
            @param src: File or directory with anisedoc sources.
            """
            if os.path.isdir(src):
                for s in os.listdir(src):
                    self.load_anisedoc_source(src + "/" + s)
            elif os.path.isfile(src):
                currentprefix = None
                currentdoc = None
                currentcontent = ""

                def end_of_doc():
                    nonlocal currentdoc, currentcontent
                    exists = False
                    for entry in self.list:
                        if entry[0] == currentdoc:
                            entry[1] += currentcontent
                            exists = True
                            break
                    if not exists:
                        self.add(currentdoc, currentcontent)
                    currentdoc = None
                    currentcontent = ""

                for line in anise.utils.basic.readfromfile(src, onestring=False):
                    if currentdoc:
                        if line.startswith(currentprefix):
                            currentcontent += line[len(currentprefix):] + "\n"
                        else:
                            end_of_doc()
                    else:
                        i = line.find("ANISEDOC ")
                        if i == 0:
                            anise.utils.logging.logwarning(
                                "skipped an ANISEDOC tag because it has a null-length prefix")
                        elif i > -1:
                            currentprefix = line[0:i]
                            if not re.search("[A-Za-z0-9]", currentprefix):
                                currentdoc = line[i + 9:].strip()
                if currentdoc:
                    end_of_doc()

    class Exports:
        """!
        Storage of all documentation export definitions (i.e. where and how documentation products are to be created).
        """

        def __init__(self):
            self.list = []

        def getbyname(self, name):
            """!
            Returns all export definitions with a given name.
            @param name: A export definition name.
            @return: List of export definitions.
            """
            return [x for x in self.list if x.name == name]

        def add(self, export):
            """!
            Exports a documentation to some targets.
            @param export: A export definition which specify an documentation export target (see Export).
            """
            self.list.append(export)

    class ShortSnippets:
        """!
        Storage of text snippets for use in the documentation texts like `@a_foo`.
        """

        def __init__(self):
            self.storage = {}
            self.add("dooverride",
                     "Override this method in custom subclasses.")
            self.add("dooverride_optional",
                     "Override this method in custom subclasses or leave the default implementation.")
            self.add("infrastructure",
                     "This is part of a particular piece of internal infrastructure and is typically not required "
                     "to be used.")

        def add(self, name, text):
            """!
            Adds a text snippet to the storage, so it can be used.
            @param name: The snippet name. This is the `foo` in `@a_foo`.
            @param text: The snippet text.
            """
            self.storage[name] = text

        def toaliases(self):
            ast = "@par\n<div class='anisedochint'>{}</div>\n\n"
            return {"a_"+k: ast.format(self.storage[k].replace("&", "&amp;").replace("<", "&lt;").
                                       replace(">", "&gt;")) for k in self.storage}

    @staticmethod
    def writedocumentations(destdir, throw_on_existing, fortype):
        """!
        Immediately exports the documentation as the export definitions currently specify.
        """
        doclist = [x for x in universe.documentation.exports.list if isinstance(x, fortype)]
        scopelist = [universe.executionscope.newsubscope(
            "Generating {} documentation ".format(y.format) + ("'{}'".format(y.heading)
                                                               if y.heading else y.key)) for y in doclist]
        for i, echan in enumerate(doclist):
            universe.executionscope.topscope().setprogress(
                anise.framework.report.ExecutionScope.ProgressDeterminate(float(i)/len(doclist)))
            with scopelist[i]:
                fullfn = destdir + "/" + echan.outputfile
                if throw_on_existing and os.path.exists(fullfn):
                    raise anise.framework.exceptions.BadFormatError(
                        "there already is a '{fullfn}', which is forbidden.".format(fullfn=fullfn))
                anise.utils.basic.mkdirs(os.path.dirname(fullfn))
                echan.generate(fullfn)

    class Export:
        """!
        Defines a documentation export.
        """

        def __init__(self, key, format, name=None, heading=None, includedeveloperdoc=False, extractall=True,
                     extractprivate=True, hideundocumented=False, imagepaths=None, additionaltext="", shownavbar=False,
                     excludepatterns=None, predefineddocmacros=None, head3=None, head4=None, source=None, sources=None):
            """!
            @param key: One of the keys stored in pool. Defines which documentation to export.
            @param format: The output format (one of documentation.ExportFormat).
            @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
            @param heading: The title.
            @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work 
                                        for all output formats).
            @param extractall: If a developer documentation should include all elements (not just the documented ones).
            @param extractprivate: If a developer documentation should include also private members.
            @param hideundocumented: If the develo documentation should entirely leave out ondocumented members.
            @param imagepaths: Directories from where images are used in the documentation.
            @param additionaltext: Additional text to append to the selected documentation.
            @param shownavbar: If to show navigation bar. Does not work in all output formats.
            @param excludepatterns: File patterns to exclude from reading.
            @param predefineddocmacros: List of additional macros to define for reading.
            @param head3: 3rd heading.
            @param head4: 4th heading.
            @param source: The documentation source root.
            @param sources: Optional list of source paths (default: `["/"]`).
            """
            imagepaths = imagepaths or []
            excludepatterns = excludepatterns or []
            predefineddocmacros = predefineddocmacros or []
            self.key = key
            self.format = format
            self.name = name
            self.heading = heading
            self.includedeveloperdoc = includedeveloperdoc
            self.extractall = extractall
            self.extractprivate = extractprivate
            self.hideundocumented = hideundocumented
            self.imagepaths = imagepaths
            self.additionaltext = additionaltext
            self.shownavbar = shownavbar
            self.excludepatterns = excludepatterns
            self.predefineddocmacros = predefineddocmacros
            self.head3 = head3
            self.head4 = head4
            self.source = source or anise.framework.files.TaskExecution(universe.packages.makerawpackage)
            self.sources = sources

        def _head3(self):
            if self.head3:
                return self.head3
            else:
                v = universe.packages.getversion()
                if v is None:
                    return ""
                else:
                    return "based on release: " + v

        def _head4(self):
            return self.head4 if (self.head4 is not None) else ""

        def generate(self, targetpath):
            """!
            Generates the content.
            @a_dooverride_optional
            """
            if self.format == ExportFormat.PDF:
                internals.generate_doxygen_userdocumentation_aspdf(
                    head1=universe.name, additionaltext=self.additionaltext, head2=self.heading,
                    sourcedir=self.source.path(), targetfile=targetpath,
                    documentations=universe.documentation.pool.gettexts(), targetdocumentation=self.key,
                    includedeveloperdoc=self.includedeveloperdoc, excludepatterns=self.excludepatterns,
                    extractprivate=self.extractprivate, hideundocumented=self.hideundocumented,
                    extractall=self.extractall, predefined=self.predefineddocmacros, imagepaths=self.imagepaths,
                    aliases=universe.documentation.snippets.toaliases(), sources=self.sources)
            elif self.format == ExportFormat.Plaintext:
                internals.generate_doxygen_userdocumentation_asplaintext(
                    projectname=universe.name, additionaltext=self.additionaltext, head=self.heading,
                    sourcedir=self.source.path(), targetfile=targetpath,
                    documentations=universe.documentation.pool.gettexts(), targetdocumentation=self.key,
                    excludepatterns=self.excludepatterns, predefined=self.predefineddocmacros,
                    aliases=universe.documentation.snippets.toaliases(), sources=self.sources)
            elif self.format == ExportFormat.HTML:
                bannerimage = universe.base.getbannerimage()
                internals.generate_doxygen_userdocumentation_ashtml(
                    head1=universe.name, head2=self.heading, head3=self._head3(), head4=self._head4(),
                    additionaltext=self.additionaltext, sourcedir=self.source.path(), targetdir=targetpath,
                    basecolor=universe.base.basecolor, bannerimage=bannerimage, showdoxygennavbar=self.shownavbar,
                    documentations=universe.documentation.pool.gettexts(),
                    anisedatadir=universe.specialpaths.anisedatadir, excludepatterns=self.excludepatterns,
                    predefined=self.predefineddocmacros, aliases=universe.documentation.snippets.toaliases(),
                    sources=self.sources)
            else:
                raise anise.framework.exceptions.BadInputError(
                    "unknown document output format '{format}'.".format(format=self.format))

    class AbstractFileExport(Export):
        """!
        Abstract base class for exports, which write documentation content to files.
        """

        def __init__(self, key, outputfile, format, name=None, heading=None, includedeveloperdoc=False,
                     extractall=True, extractprivate=True, hideundocumented=False, imagepaths=None, additionaltext="",
                     shownavbar=False, excludepatterns=None, predefineddocmacros=None, head3=None, head4=None,
                     source=None, sources=None):
            internals.Export.__init__(self, key, format, name, heading, includedeveloperdoc, extractall, extractprivate,
                                      hideundocumented, imagepaths, additionaltext, shownavbar, excludepatterns,
                                      predefineddocmacros, head3, head4, source, sources)
            self.outputfile = outputfile

    @staticmethod
    def loadfile(name):
        return anise.utils.basic.readfromfile(os.path.dirname(__file__) + "/documentation/" + name)

    @staticmethod
    def scalecolor(c, brightness, saturation=None):
        """!
        Internally used for scaling a color tuple to a new one with the same hue but a given brightness.
        """
        h, l, s = colorsys.rgb_to_hls(*c)
        if saturation is not None:
            s *= saturation
            if s > 1:
                s = 1.0
        return colorsys.hls_to_rgb(h, brightness, s)

    @staticmethod
    def derivecolors(c, saturation=None):
        """!
        Internally used for computing some css colors for the given basecolor (with different brightnesses).
        """
        scale = internals.scalecolor
        return ["rgb({0},{1},{2})".format(*[min(int(x * 255), 255) for x in _c]) for _c in [
            scale(c, 0.0, saturation), scale(c, 0.1, saturation), scale(c, 0.2, saturation), scale(c, 0.3, saturation),
            scale(c, 0.4, saturation), scale(c, 0.5, saturation), scale(c, 0.6, saturation), scale(c, 0.7, saturation),
            scale(c, 0.8, saturation), scale(c, 0.9, saturation),
        ]]

    @staticmethod
    def generate_doxygen_userdocumentation_ashtml(
            *, head1, head2, head3="", head4="", additionaltext="", sourcedir, targetdir, basecolor, bannerimage=None,
            documentations, showdoxygennavbar=False, anisedatadir, doxygensettings="", excludepatterns=None,
            predefined=None, aliases=None, sources=None, mergehead1and2totitle=False):
        """!
        Writes all user documentations as html (powered by Doxygen). You find it afterwards in the `(documentkey).html`
        files (and tons of other files).
        This will contain the developer documentation as well, so references into it will work.
        @param head1: The first heading (typically the project name).
        @param head2: The second heading.
        @param head3: The third heading.
        @param head4: The fourth heading.
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the html output.
        @param basecolor: The basecolor for colouring the output.
        @param bannerimage: Absolute path to the banner image (will be copied).
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param showdoxygennavbar: If the doxygen navigation bar should be visible.
        @param anisedatadir: Path to the Anise data directory (which contains some data files which might be needed).
        @param doxygensettings: Additional Doxygen configuration for finetuning.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param predefined: A list of (C++) macros to define before the source files are parsed.
        @param aliases: Custom command aliases (as string/string-dict).
        @param sources: Optional list of source paths (default: `["/"]`).
        @param mergehead1and2totitle: If head1 and head2 shall be merged together for the page title.
        """
        pnamesettings = 'PROJECT_NAME="{0}"\n'.format((head1 + ": " + head2)
                                                      .replace('"', r'\"')) if mergehead1and2totitle else ""
        internals.generate_doxygen_documentation(
            head1, head2, head3, head4, sourcedir, targetdir, pnamesettings + doxygensettings, "HTML", additionaltext,
            basecolor, bannerimage, [], showdoxygennavbar, internals.loadfile("__doxygen_userdocumentation.css"),
            documentations, anisedatadir=anisedatadir, htmlheadertranslation=lambda x: x.replace("$title", head2 or ""),
            excludepatterns=excludepatterns, predefined=predefined, aliases=aliases, sources=sources)

    @staticmethod
    def generate_doxygen_userdocumentation_asmanpage(
            *, projectname, entityname, section, additionaltext="", sourcedir, targetdir, documentations,
            targetdocumentation, doxygensettings="", excludepatterns=None, predefined=None, aliases=None, sources=None):
        """!
        Writes one user documentation page as manpage (powered by Doxygen) and returns the path to the result file.
        @param projectname: The project name.
        @param entityname: The `NAME` for the manpage (typically the name of the command, function or whatever).
        @param section: The section number (see man manpage).
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the manpage output.
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param targetdocumentation: The key for the documentation which must be written.
        @param doxygensettings: Additional Doxygen configuration for finetuning.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param predefined: A list of (C++) macros to define before the source files are parsed.
        @param aliases: Custom command aliases (as string/string-dict).
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        entityname = entityname.replace(" ", "-")
        _itext = """/*!
\\page {entityname}
\\copydoc {targetdocumentation}
        */
        """.format(**locals())
        internals.generate_doxygen_documentation(
            projectname, None, None, None, sourcedir, targetdir,
            'MAN_EXTENSION=".{0}"\n'.format(section) + doxygensettings, "MAN", additionaltext + "\n" + _itext,
            (0, 0, 7), None, [], False, "", documentations, excludepatterns=excludepatterns, predefined=predefined,
            aliases=aliases, sources=sources)
        return "{targetdir}/man/man{section}/{entityname}.{section}".format(**locals())

    @staticmethod
    def generate_doxygen_userdocumentation_asplaintext(
            *, projectname, head, additionaltext="", sourcedir, targetfile, documentations, targetdocumentation,
            doxygensettings="", excludepatterns=None, predefined=None, aliases=None, sources=None):
        """!
        Writes one user documentation page as plain text (powered by Doxygen).
        @param projectname: The project name.
        @param head: The heading.
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetfile: The target file for the plaintext output.
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param targetdocumentation: The key for the documentation which must be written.
        @param doxygensettings: Additional Doxygen configuration for finetuning.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param predefined: A list of (C++) macros to define before the source files are parsed.
        @param aliases: Custom command aliases (as string/string-dict).
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        with anise.framework.files.TempDir() as tempdir:
            f = internals.generate_doxygen_userdocumentation_asmanpage(
                projectname=projectname, entityname=head, section=7, additionaltext=additionaltext, sourcedir=sourcedir,
                targetdir=tempdir.path(), documentations=documentations, targetdocumentation=targetdocumentation,
                doxygensettings=doxygensettings, excludepatterns=excludepatterns, predefined=predefined,
                aliases=aliases, sources=sources)
            r, plaintext = anise.utils.basic.call(
                "man -l {0} | col -b".format(shlex.quote(f)), shell=True,
                raise_on_errors="Problem while unformatting man page")
            res = ["  " + projectname + ": " + head]
            res.append("-" * len(res[0]) + "\n")
            began = None
            for l in plaintext.split("\n"):
                if began is not None and began >= 0:
                    res.append(l)
                elif began is None and l.startswith("NAME"):
                    began = -1
                elif began is not None:
                    began += 1
        anise.utils.basic.writetofile(targetfile, "\n".join(res[:-1]))

    @staticmethod
    def generate_doxygen_userdocumentation_aspdf(
            *, head1, head2, additionaltext="", sourcedir, targetfile, documentations, targetdocumentation,
            doxygensettings="", includedeveloperdoc=False, excludepatterns=None, extractprivate=True, extractall=True,
            hideundocumented=False, predefined=None, imagepaths=None, aliases=None, sources=None):
        """!
        Writes one user documentation page as pdf (powered by Doxygen).
        @param head1: The first heading (typically the project name).
        @param head2: The second heading.
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetfile: The target file for the pdf output.
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param targetdocumentation: The key for the documentation which must be written.
        @param doxygensettings: Additional Doxygen configuration for finetuning.
        @param includedeveloperdoc: If the pdf shall include the developer documentation.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param extractprivate: If private members shall be extracted to the developer documentation.
        @param extractall: If all members shall be extracted to the developer documentation.
        @param hideundocumented: If undocumented member shall be left out in the developer documentation.
        @param predefined: A list of (C++) macros to define before the source files are parsed.
        @param imagepaths: Directories from where images are used in the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        with anise.framework.files.TempDir() as tempdir:
            internals.generate_doxygen_documentation(
                head1, head2, None, None, sourcedir, tempdir.path(), "LATEX_BATCHMODE=YES\n" + doxygensettings, "LATEX",
                additionaltext, (0, 0, 7), None, [], False, "", documentations, excludepatterns=excludepatterns,
                extractprivate=extractprivate, extractall=extractall, hideundocumented=hideundocumented,
                predefined=predefined, imagepaths=imagepaths, aliases=aliases, sources=sources)
            refmantex = anise.utils.basic.readfromfile(tempdir.path() + "/latex/refman.tex", onestring=False)
            refmantexnew = []
            intitlepage = False
            chapteraccepted = False
            for l in refmantex:
                if intitlepage:
                    if l.startswith("\\end{titlepage"):
                        intitlepage = False
                    continue
                elif l.startswith("\\begin{titlepage"):
                    intitlepage = True
                    continue
                elif l.startswith("\\chapter{"):
                    if not includedeveloperdoc or (l[9:l.find("}")] not in ["Namespace Index", "Hierarchical Index",
                                                                            "Class Index", "Namespace Documentation",
                                                                            "Class Documentation"]):
                        chapteraccepted = False
                        continue
                    chapteraccepted = True
                elif l.startswith("\\input{"):
                    if l.find("{" + targetdocumentation + "}") > -1:
                        refmantexnew.append(
                            "\\chapter{" + head2 + "}\n\\label{" + head2 + "}\n\\hypertarget{" + targetdocumentation +
                            "}{}\n\\input{" + targetdocumentation + "}\n")
                    elif chapteraccepted:
                        refmantexnew.append(l)
                    else:
                        continue
                elif l.startswith("\\label{"):
                    continue
                elif l.startswith("\\hypertarget{"):
                    continue
                else:
                    refmantexnew.append(l)
            anise.utils.basic.writetofile(tempdir.path() + "/latex/refman.tex", "".join(refmantexnew))
            tex = anise.utils.basic.readfromfile(tempdir.path() + "/latex/" + targetdocumentation + ".tex",
                                                 onestring=False)
            texstring = "".join(tex)
            texnew = []
            for l in tex:
                ll = ""
                iend = 0
                if not includedeveloperdoc:
                    for linkmatch in re.finditer(r"\\hyperlink\{([^{}]+)\}\{([^{}]+)\}", l):
                        starthl = linkmatch.start(0)
                        endhl = linkmatch.end(0)
                        linktgt = linkmatch.group(1)
                        linklbl = linkmatch.group(2)
                        exists = texstring.find("\\hypertarget{" + linktgt + "}") > -1
                        if exists:
                            ll += l[iend:endhl]
                        else:
                            ll += l[iend:starthl] + linklbl
                        iend = endhl
                    l = ll + l[iend:]
                if re.match("^\\\\[a-z]*section\\*\\{", l):
                    ast = l.find("*")
                    texnew.append(l[0:ast] + l[ast + 1:])
                else:
                    texnew.append(l)
            anise.utils.basic.writetofile(tempdir.path() + "/latex/" + targetdocumentation + ".tex", "".join(texnew))
            with anise.utils.basic.ChDir(tempdir.path() + "/latex"):
                r, o = anise.utils.basic.call("make", "refman.pdf")
                anise.utils.logging.logdebug(o)
            tempdir.dl(subpath="latex/refman.pdf", to=targetfile)

    @staticmethod
    def generate_doxygen_homepage(
            *, head1, head2, head3, head4, sourcedir, targetdir, basecolor, bannerimage, sections, anisedatadir,
            doxygensettings="", excludepatterns=None, predefined=None, aliases=None, sources=None):
        """!
        Writes the project homepage (powered by Doxygen). You find it afterwards in `index.html`
        (and tons of other files).
        @param head1: The first heading (typically the project name).
        @param head2: The second heading (typically a one-liner description).
        @param head3: The third heading (typically version information).
        @param head4: The fourth heading.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the html output.
        @param basecolor: The basecolor for colouring the output.
        @param bannerimage: Absolute path to the banner image (will be copied).
        @param sections: A list of (sectiontitle,sectionsource,additionalparams) tuples for the actual content.
        @param anisedatadir: Absolute path to the anise data dir (needed for finding icons, ...; typically somewhere
                             in the anise installation path).
        @param doxygensettings: Additional Doxygen configuration for finetuning.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param predefined: A list of (C++) macros to define before the source files are parsed.
        @param aliases: Custom command aliases (as string/string-dict).
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        page = "/*!\n\\mainpage " + head1 + "\n"
        headlinks = []
        for sectionkey, sectiontext, additionalparams, sectionhead in sections:
            if sectionhead == "icon":
                headlinks.append((sectionhead, sectiontext))
            else:
                sectionanchor = "".join([x for x in sectionkey.lower() if x in "abcdefghijklmnopqrstuvwxyz0123456789"])
                if additionalparams.get("important", False):
                    sectionanchor += "_i_"
                page += "\\anchor {sectionanchor}\n# {sectionhead} #\n{sectiontext}\n".format(**locals())
                headlinks.append((sectionhead, "#" + sectionanchor))
        page += "*/"
        internals.generate_doxygen_documentation(
            head1, head2, head3, head4, sourcedir, targetdir, "PROJECT_NAME=\n" + doxygensettings, "HTML", page,
            basecolor, bannerimage, headlinks, False, internals.loadfile("__doxygen_homepage.css"), [],
            anisedatadir=anisedatadir, excludepatterns=excludepatterns, predefined=predefined, aliases=aliases,
            sources=sources)
        return targetdir

    @staticmethod
    def generate_doxygen_documentation(
            head1, head2, head3, head4, sourcedir, targetdir, doxygensettings, outputformat, additionalsources,
            basecolor, bannerimage, headlinks, showdoxygennavbar, css, documentations, anisedatadir=None,
            htmlheadertranslation=None, extractprivate=True, extractall=True, hideundocumented=False,
            excludepatterns=None, predefined=None, imagepaths=None, aliases=None, sources=None):
        """!
        Writes some output with Doxygen. Mostly used by some higher-level functions in this module.
        @param head1: The first heading.
        @param head2: The second heading.
        @param head3: The third heading.
        @param head4: The fourth heading.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the output.
        @param doxygensettings: Additional Doxygen configuration for finetuning.
        @param outputformat: The desired output format (as in doxygen config's GENERATE_xxx keys).
        @param additionalsources: Auxiliary source code passed to Doxygen for injecting additional content.
        @param basecolor: The basecolor for colouring the output.
        @param bannerimage: Absolute path to the banner image (will be copied).
        @param headlinks: Links in the heading panel (html output only).
        @param showdoxygennavbar: If the doxygen navigation bar should be visible.
        @param css: Additional css (html output only).
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param anisedatadir: Absolute path to the anise data dir (needed for finding icons, ...; typically somewhere
                             in the anise installation path).
        @param htmlheadertranslation: A function which can replace stuff in the default html header (html output only).
        @param extractprivate: If private members shall be extracted to the developer documentation.
        @param extractall: If all members shall be extracted to the developer documentation.
        @param hideundocumented: If undocumented member shall be left out in the developer documentation.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param predefined: A list of (C++) macros to define before the source files are parsed.
        @param imagepaths: Directories from where images are used in the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        excludepatterns = excludepatterns or []
        predefined = predefined or []
        imagepaths = imagepaths or []
        aliases = aliases or {}
        sources = sources or ["/"]
        if callable(additionalsources):
            additionalsources = additionalsources()
        imagepaths = list(imagepaths)
        imagepaths.append(sourcedir)
        if anisedatadir:
            imagepaths.append(anisedatadir)
        simagepath = " ".join(['"' + x + '"' for x in imagepaths])
        if bannerimage:
            shutil.copyfile(bannerimage, targetdir + "/banner.png")
            bannerimage = "banner.png"
        if not head2:
            head2 = ""
        if not head3:
            head3 = ""
        if not head4:
            head4 = ""
        col00, col01, col02, col03, col04, col05, col06, col07, col08, col09 = internals.derivecolors(basecolor)
        colS00, colS01, colS02, colS03, colS04, colS05, colS06, colS07, colS08, colS09 = internals.derivecolors(
            basecolor, 1.3)
        _lb = "{"
        _rb = "}"
        additionalcss = (internals.loadfile("__doxygen_main.css") + css).replace("}", "##LB_##") \
                         .replace("{", "{_lb}").replace("##LB_##", "{_rb}").replace(">", "}").replace("<", "{").format(
            **locals()) + "\n"
        if not showdoxygennavbar:
            additionalcss += """\n#navrow1, #navrow2, #navrow3, #navrow4, #navrow5, 
            #navrow6, #navrow7, #navrow8, #navrow9, #main-nav {
            visibility: collapse;}\n"""

        r, g, b = (x * 255 for x in basecolor)
        colorhue = int(math.atan2(1.732050808 * (g - b), (2 * r - g - b)) * 57.295779513) % 360
        hheadlinks = ""
        if headlinks:
            for hname, hval in headlinks:
                if hname == "icon":
                    tfile = "icon_" + os.path.basename(hval)
                    shutil.copyfile(hval, targetdir + "/" + tfile)
                    alt = os.path.basename(hval)
                    if "." in alt:
                        alt = alt[:alt.find(".")]
                    hheadlinks += " <img src='{tfile}' alt='{alt}' title='{alt}' class='sectionicon' /> ".format(
                        **locals())
                else:
                    hheadlinks += " <a class='sectionlink indexlevel1 indexlevel2' href='{hval}'>{hname}</a> " \
                        .format(**locals())
        bannerimg = "<img id='bannerimage' src='{0}'/>".format(bannerimage) if bannerimage else ""
        htmlcontent = internals.loadfile("__doxygen_main.html")
        htmlsplitter = "<!-- INNER"
        ihtmlsplit = htmlcontent.find(htmlsplitter)
        if ihtmlsplit == -1:
            raise anise.framework.exceptions.UnexpectedSituationError("doxygen_main html malformed: splitter not found")
        htmlheader = htmlcontent[:ihtmlsplit]
        if htmlheadertranslation:
            htmlheader = htmlheadertranslation(htmlheader)
        htmlheaderjavascript = internals.loadfile("__doxygen_main.js")
        htmlheader = htmlheader.format(**locals())
        for dockey, docsrc in documentations:
            additionalsources += "\n/*!\n\\page " + dockey + "\n" + docsrc + "\n*/\n"
        uoformat = outputformat.upper()
        spredefined = " \\\n".join(['"{}"'.format(x) for x in predefined])
        sexcludedpatterns = " \\\n".join(['"{}"'.format(x) for x in excludepatterns])
        sextractall = "YES" if extractall else "NO"
        sextractprivate = "YES" if extractprivate else "NO"
        shideundoc = "YES" if hideundocumented else "NO"
        saliases = ""
        for kalias in aliases:
            valias = aliases[kalias].replace("\\", "\\\\").replace("\"", "\\\"").replace("\n", "\\n")
            saliases += "ALIASES += \"{}={}\"\n".format(kalias, valias)
        ssources = " ".join(['"'+sourcedir+"/"+x+'"' for x in sources])
        doxygen_baseconfig = internals.loadfile("__doxygen_baseconfig.ini")
        with anise.framework.files.TempDir() as tempdir:
            tpath = tempdir.path()
            anise.utils.basic.writetofile(tpath + "/_anise_additionalstuff.cpp", additionalsources)
            anise.utils.basic.writetofile(tpath + "/_anise_htmlheader", htmlheader)
            anise.utils.basic.writetofile(tpath + "/_anise_htmlfooter", htmlcontent[ihtmlsplit:])
            anise.utils.basic.writetofile(tpath + "/_anise_additionalcss.css", additionalcss)
            anise.utils.basic.writetofile(tpath + "/cfg",
                                          doxygen_baseconfig.format(**locals()) + "\n" + saliases + doxygensettings)
            r, o = anise.utils.basic.call("doxygen", tpath + "/cfg",
                                          raise_on_errors="unable to generate content via doxygen")
            anise.utils.logging.logdebug(o)


class ExportFormat:
    """!
    Enumeration of output formats.
    """
    Plaintext = "Plaintext"
    PDF = "PDF"
    HTML = "HTML"


class HomepageExport(internals.Export):
    """!
    Describes an export of an anise documentation directly into the homepage.
    """

    def __init__(self, key, name=None, heading=None, includedeveloperdoc=False, extractall=True, extractprivate=True,
                 hideundocumented=False, imagepaths=None, additionaltext="", shownavbar=False, excludepatterns=None,
                 predefineddocmacros=None, head3=None, head4=None, source=None, sources=None, linktarget=None,
                 index=100000):
        """!
        @param key: One of the keys stored in pool. Defines which documentation to export.
        @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
        @param heading: The title.
        @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work for
                                    all output formats).
        @param extractall: If a developer documentation should include all elements (not just the documented ones).
        @param extractprivate: If a developer documentation should include also private members.
        @param hideundocumented: If the develo documentation should entirely leave out ondocumented members.
        @param imagepaths: Directories from where images are used in the documentation.
        @param additionaltext: Additional text to append to the selected documentation.
        @param shownavbar: If to show navigation bar. Does not work in all output formats.
        @param excludepatterns: File patterns to exclude from reading.
        @param predefineddocmacros: List of additional macros to define for reading.
        @param head3: 3rd heading.
        @param head4: 4th heading.
        @param source: The documentation source root.
        @param sources: Optional list of source paths (default: `["/"]`).
        @param linktarget: Link target.
        @param index: A position number. Used for ordering.
        """
        internals.Export.__init__(self, key, ExportFormat.HTML, name, heading, includedeveloperdoc, extractall,
                                  extractprivate, hideundocumented, imagepaths, additionaltext, shownavbar,
                                  excludepatterns, predefineddocmacros, head3, head4, source, sources)
        self.linktarget = linktarget
        self.index = index


class PackageExport(internals.AbstractFileExport):
    """!
    Describes an export of an anise documentation into the distribution package(s).
    """

    def __init__(self, key, outputfile, format, name=None, heading=None, includedeveloperdoc=False, extractall=True,
                 extractprivate=True, hideundocumented=False, imagepaths=None, additionaltext="", shownavbar=False,
                 excludepatterns=None, predefineddocmacros=None, head3=None, head4=None, source=None, sources=None):
        """!
        @param key: One of the keys stored in pool. Defines which documentation to export.
        @param outputfile: Target file path relative to the 'raw package's root.
        @param format: The output format (one of documentation.ExportFormat).
        @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
        @param heading: The title.
        @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work for
                                    all output formats).
        @param extractall: If a developer documentation should include all elements (not just the documented ones).
        @param extractprivate: If a developer documentation should include also private members.
        @param hideundocumented: If the develo documentation should entirely leave out ondocumented members.
        @param imagepaths: Directories from where images are used in the documentation.
        @param additionaltext: Additional text to append to the selected documentation.
        @param shownavbar: If to show navigation bar. Does not work in all output formats.
        @param excludepatterns: File patterns to exclude from reading.
        @param predefineddocmacros: List of additional macros to define for reading.
        @param head3: 3rd heading.
        @param head4: 4th heading.
        @param source: The documentation source root.
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        internals.AbstractFileExport.__init__(self, key, outputfile, format, name, heading, includedeveloperdoc,
                                              extractall, extractprivate, hideundocumented, imagepaths, additionaltext,
                                              shownavbar, excludepatterns, predefineddocmacros, head3, head4, source,
                                              sources)


class SourceExport(internals.AbstractFileExport):
    """!
    Describes an export of an anise documentation back into the source tree.
    """

    def __init__(self, key, outputfile, format, name=None, heading=None, includedeveloperdoc=False, extractall=True,
                 extractprivate=True, hideundocumented=False, imagepaths=None, additionaltext="", shownavbar=False,
                 excludepatterns=None, predefineddocmacros=None, head3=None, head4=None, source=None, sources=None):
        """!
        @param key: One of the keys stored in pool. Defines which documentation to export.
        @param outputfile: Target file path relative to the 'raw package's root.
        @param format: The output format (one of documentation.ExportFormat).
        @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
        @param heading: The title.
        @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work for
                                    all output formats).
        @param extractall: If a developer documentation should include all elements (not just the documented ones).
        @param extractprivate: If a developer documentation should include also private members.
        @param hideundocumented: If the develo documentation should entirely leave out ondocumented members.
        @param imagepaths: Directories from where images are used in the documentation.
        @param additionaltext: Additional text to append to the selected documentation.
        @param shownavbar: If to show navigation bar. Does not work in all output formats.
        @param excludepatterns: File patterns to exclude from reading.
        @param predefineddocmacros: List of additional macros to define for reading.
        @param head3: 3rd heading.
        @param head4: 4th heading.
        @param source: The documentation source root.
        @param sources: Optional list of source paths (default: `["/"]`).
        """
        internals.AbstractFileExport.__init__(self, key, outputfile, format, name, heading, includedeveloperdoc,
                                              extractall, extractprivate, hideundocumented, imagepaths, additionaltext,
                                              shownavbar, excludepatterns, predefineddocmacros, head3, head4, source,
                                              sources)


def has_doxygen_developerdocumentation(also_include_in_readme=False, **kwargs):
    """!
    Configures a project for creating a Doxygen developer documentation.
    @param also_include_in_readme: If a developer documentation shall be included in the README files.
    @param kwargs: Additional arguments for the export definition.
    """
    def _add_homepage_export():
        additionaltext = lambda: """
    /**
\mainpage Main page

This is a documentation for developers of {projectname} as well as interested hackers, who want
to work with the source code. It is automatically created from the source code comments. It is
not guaranteed that the information is correct or up-to-date (even if it is based on the latest version).
However, it should give a good overview.

You are current reading the developer documentation based on version {version}.

{maturity}
    */""".format(projectname=universe.name, version=universe.packages.getversion(),
                 maturity=universe.base.getprojectstatusdescription())
        universe.documentation.exports.add(
            HomepageExport(key="developerdoc", heading="Developer Documentation", includedeveloperdoc=True,
                           additionaltext=additionaltext, shownavbar=True, linktarget="index.html", **kwargs))
    universe.addhook(anise.framework.features.HOOK_BEFORE_EXECUTION, _add_homepage_export)
    universe.documentation.readme_include_developer_doc = kwargs if also_include_in_readme else False


def markdown_downgrade_headings(docsrc, levels):
    """!
    Rewrites a Markdown documentation for shifting the heading levels down.
    @param docsrc: The Markdown source as string.
    @param levels: Number of levels to downshift the headings.
    @return: The result Markdown documentation.
    """
    result = ""
    moresharps = "#" * levels
    for line in docsrc.split("\n"):
        if line.startswith("#"):
            line = moresharps + line
        result += line + "\n"
    return result[:-1]
