# Copyright (C) 2014-2019, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for helpers related to web-based user interfaces.
"""

import json
import os
import time

import anise.data.PyWebApp.pywebapp
import anise.framework.engine
import anise.framework.features
import anise.framework.globalvars
import anise.framework.report
import anise.utils

from anise.framework.projects import universe

uiweb = anise.framework.features.loadfeature("ui.web").featuremodule


class AniseWebApplication(anise.data.PyWebApp.pywebapp.Application):

    def __init__(self, title, parentid=None, returntoparent=True):
        anise.data.PyWebApp.pywebapp.Application.__init__(self, parentid=parentid, returntoparent=returntoparent)
        self.startmessageonlyterminal = False
        self.seticon("static/icons/pde/anise.png")
        self.sethead1(title)
        self._wh_hooks = []
        self.userfeedbackanswers = {}
        self.userfeedbackrequest = None
        self.addclientscript(frompath=os.path.dirname(os.path.abspath(__file__))+"/__helpers.js")
        self.addclientstyle(frompath=os.path.dirname(os.path.abspath(__file__))+"/__helpers.css")

    def _execution_scopepart_added(self, x):
        self.triggerevent("anise_ExecPager_newcontentavailable")

    def _send_userfeedback_web_request(self, request):
        self.userfeedbackrequest = request
        self.triggerevent("anise_UserFeedback_newcontentavailable")

    def _get_userfeedback_web_answers(self, requestindex):
        result = self.userfeedbackanswers.get(requestindex)
        if result is not None:
            self.userfeedbackanswers.pop(requestindex)
        return result

    def _shutdown(self):
        time.sleep(3)
        self.stop()

    def oninitialize(self):
        self.appconfig.setconfig("aniseProjectName", universe.name)
        self.appconfig.setconfig("aniseShowLogDebug", bool(anise.framework.globalvars.logdebug))
        self._pyhtmlview_additionalargs = ["--icon", universe.specialpaths.anisedatadir + "/icons/pde/anise.png"]

    def start(self):
        self._wh_hooks = [universe.addhook(anise.framework.report.HOOK_EXECUTION_SCOPEPART_ADDED,
                                           self._execution_scopepart_added),
                          universe.addhook(uiweb.HOOK_SEND_USERFEEDBACK_WEB_REQUEST,
                                           self._send_userfeedback_web_request),
                          universe.addhook(uiweb.HOOK_GET_USERFEEDBACK_WEB_ANSWERS,
                                           self._get_userfeedback_web_answers),
                          universe.addhook(anise.framework.engine.HOOK_SHUTDOWN, self._shutdown, provides=["ui.web"])]
        anise.data.PyWebApp.pywebapp.Application.start(self)

    def stop(self):
        anise.data.PyWebApp.pywebapp.Application.stop(self)
        for h in self._wh_hooks:
            universe.removehook(h)
        self._wh_hooks = []

    def onprocessrequesterror(self, request, error):
        anise.utils.logging.logexception(error, prolog="Error in Anise web application: ")

    def onopenbrowsererror(self):
        anise.utils.logging.logerror("There were problems while starting a web browser for controlling Anise\n"
                                     "execution.\n\n"
                                     "Please try to open it manually or start Anise again\n"
                                     "with the '--no-gui' parameter added or read the manual.", printcaller=False)

    def onbrowserreopened(self):
        anise.utils.logging.logwarning("Your Anise browser session was closed prematurely.\nA new one is starting.",
                                       printcaller=False)

    def onopenbrowserinformationoutput(self, kind, message):
        anise.utils.logging.loginfo(message, printcaller=False, terminalonly=True)

    def clientstaticfilelocations(self):
        return [universe.specialpaths.anisedatadir, ]

    @anise.data.PyWebApp.pywebapp.do_with_param_type("since", int)
    def do_getprogresses(self, request, since):
        items = []
        progresses = []
        for sp in anise.framework.report.rootExecutionScope.getscopepartsbyidrange():
            if isinstance(sp, anise.framework.report.ExecutionScope) \
                    and not isinstance(sp.progress(), anise.framework.report.ExecutionScope.ProgressFinished):
                progresses.append((sp.address(), sp.progress().tojson()))
            if sp.id() > since:
                if isinstance(sp, anise.framework.report.ExecutionScope):
                    item = {"title": sp.title()}
                else:
                    item = {"value": sp.value(), "hints": sp.hints()}
                item.update({"id": sp.id(), "address": sp.address()})
                items.append(item)
        return {'items': items, 'progresses': progresses,
                'currentscope': anise.framework.report.rootExecutionScope.topscope().address()}

    def do_getuserfeedback(self, request):
        if self.userfeedbackrequest and self.userfeedbackrequest.isactive():
            return self.userfeedbackrequest.tojson()
        else:
            return None

    @anise.data.PyWebApp.pywebapp.do_with_param_type("index", int)
    def do_answeruserfeedback(self, request, index, answer):
        self.userfeedbackanswers[index] = json.loads(answer)
