# Copyright (C) 2014-2019, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for Webexec, which is the default web interface for task executions.
"""

import ctypes
import os
import threading

import anise.framework.features

from anise.framework.projects import universe

ui = anise.framework.features.loadfeature("ui").featuremodule
webhelpers = anise.framework.features.loadfeature("ui.web.helpers").featuremodule


class WebexecApplication(webhelpers.AniseWebApplication):

    class AbortedInterrupt(KeyboardInterrupt):

        def __init__(self):
            KeyboardInterrupt.__init__(self)

        def __str__(self):
            return "Aborted by user."

    def __init__(self, parentid=None, returntoparent=True):
        webhelpers.AniseWebApplication.__init__(self, "Anise Task Execution: " + universe.name, parentid=parentid,
                                                returntoparent=returntoparent)
        self.startmessageonlyterminal = True
        self.mainthread = threading.current_thread()
        self.mainexecscope = universe.executionscope.topscope()
        self._shutdownByUser = False
        self.addclientscript(frompath=os.path.dirname(os.path.abspath(__file__))+"/__webexec.js")
        self.addclientstyle(frompath=os.path.dirname(os.path.abspath(__file__))+"/__webexec.css")
        self.universe = None

    def oninitialize(self):
        super().oninitialize()
        self.universe = universe
        self.appconfig.setconfig("skipShutdownDialog", True)
        self.appconfig.setconfig("showBrowserClosedNotification", True)

    def stop(self):
        stopimplicit = self.appconfig.getconfig("stopImplicitlyWhenBrowserClosed")
        if self._shutdownByUser or stopimplicit:
            exc = ctypes.py_object(WebexecApplication.AbortedInterrupt)
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.mainthread.ident), exc)
            self.mainthread.join()
            if stopimplicit:  # just for the case anyone unset that meanwhile (now there is no way back)
                self.appconfig.setconfig("stopImplicitlyWhenBrowserClosed", True)
        super().stop()

    def do_application_shutdown(self, request):
        self._shutdownByUser = True
        return super().do_application_shutdown(request)


@anise.framework.features.hookhandler(ui.HOOK_UIMODE_SELECTED)
def initproject():
    if universe.ui.mode == ui.Mode.WEB:
        ui.web.webexec.instance = WebexecApplication(parentid=getattr(ui.web.webexec, "parentid", None),
                                                     returntoparent=getattr(ui.web.webexec, "returntoparent", True))
        class SetImplicitStopAllowed(ui.internals.SetImplicitStopAllowed):
            def __init__(self, val):
                self.val = val
            def __enter__(self):
                self.o = ui.web.webexec.instance.SetStopImplicitlyWhenBrowserClosed(ui.web.webexec.instance, self.val)
                return self.o.__enter__()
            def __exit__(self, exc_type, exc_val, exc_tb):
                return self.o.__exit__(exc_type, exc_val, exc_tb)
        universe.addhook(ui.HOOK_GET_ALLOWIMPLICITCLOSE_IMPLEMENTATION, SetImplicitStopAllowed, prepares=["ui"])
        ui.web.webexec.instance.start()
